@@ -1,12 +1,10 @@
-# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2014 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 from __future__ import absolute_import, print_function
 
-import atexit
 import contextlib
 import errno
-import fcntl
 import itertools
 import os
 import re
@@ -21,23 +19,37 @@ from collections import defaultdict, namedtuple
 from contextlib import contextmanager
 from datetime import datetime
 from uuid import uuid4
+from zipfile import ZipFile, ZipInfo
 
-from pex.typing import TYPE_CHECKING
+from pex import atexit
+from pex.enum import Enum
+from pex.executables import chmod_plus_x
+from pex.fs import safe_link, safe_rename, safe_symlink
+from pex.os import is_exe
+from pex.typing import TYPE_CHECKING, cast
 
 if TYPE_CHECKING:
     from typing import (
         Any,
         Callable,
+        Container,
         DefaultDict,
+        Dict,
         Iterable,
         Iterator,
+        List,
         NoReturn,
         Optional,
         Set,
         Sized,
+        Text,
         Tuple,
+        TypeVar,
+        Union,
     )
 
+    _Text = TypeVar("_Text", bytes, str, Text)
+
 # We use the start of MS-DOS time, which is what zipfiles use (see section 4.4.6 of
 # https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT).
 DETERMINISTIC_DATETIME = datetime(
@@ -46,27 +58,32 @@ DETERMINISTIC_DATETIME = datetime(
 _UNIX_EPOCH = datetime(year=1970, month=1, day=1, hour=0, minute=0, second=0, tzinfo=None)
 DETERMINISTIC_DATETIME_TIMESTAMP = (DETERMINISTIC_DATETIME - _UNIX_EPOCH).total_seconds()
 
-
-def filter_pyc_dirs(dirs):
-    # type: (Iterable[str]) -> Iterator[str]
-    """Return an iterator over the input `dirs` filtering out Python bytecode cache directories."""
-    for d in dirs:
-        if d != "__pycache__":
-            yield d
+# N.B.: The `SOURCE_DATE_EPOCH` env var is semi-standard magic for controlling
+# build tools. Wheel, for example, has supported this since 2016.
+# See:
+# + https://reproducible-builds.org/docs/source-date-epoch/
+# + https://github.com/pypa/wheel/blob/1b879e53fed1f179897ed47e55a68bc51df188db/wheel/archive.py#L36-L39
+REPRODUCIBLE_BUILDS_ENV = dict(
+    PYTHONHASHSEED="0", SOURCE_DATE_EPOCH=str(int(DETERMINISTIC_DATETIME_TIMESTAMP))
+)
 
 
-def filter_pyc_files(files):
-    # type: (Iterable[str]) -> Iterator[str]
-    """Iterate the input `files` filtering out any Python bytecode files."""
-    for f in files:
-        # For Python 2.7, `.pyc` files are compiled as siblings to `.py` files (there is no
-        # __pycache__ dir).
-        if not f.endswith(".pyc") and not is_pyc_temporary_file(f):
-            yield f
+def is_pyc_dir(dir_path):
+    # type: (Text) -> bool
+    """Return `True` if `dir_path` is a Python bytecode cache directory."""
+    return os.path.basename(dir_path) == "__pycache__"
+
+
+def is_pyc_file(file_path):
+    # type: (Text) -> bool
+    """Return `True` if `file_path` is a Python bytecode file."""
+    # N.B.: For Python 2.7, `.pyc` files are compiled as siblings to `.py` files (there is no
+    # __pycache__ dir).
+    return file_path.endswith((".pyc", ".pyo")) or is_pyc_temporary_file(file_path)
 
 
 def is_pyc_temporary_file(file_path):
-    # type: (str) -> bool
+    # type: (Text) -> bool
     """Check if `file` is a temporary Python bytecode file."""
     # We rely on the fact that the temporary files created by CPython have object id (integer)
     # suffixes to avoid picking up files where Python bytecode compilation is in-flight; i.e.:
@@ -81,56 +98,36 @@ def die(msg, exit_code=1):
 
 
 def pluralize(
-    subject,  # type: Sized
+    subject,  # type: Union[int, Sized]
     noun,  # type: str
 ):
     # type: (...) -> str
     if noun == "":
         return ""
-    count = len(subject)
+    count = subject if isinstance(subject, int) else len(subject)
     if count == 1:
         return noun
-    if noun[-1] in ("s", "x", "z") or noun[-2:] in ("sh", "ch"):
+    if noun[-1] == "y":
+        return noun[:-1] + "ies"
+    elif noun[-1] in ("s", "x", "z") or noun[-2:] in ("sh", "ch"):
         return noun + "es"
     else:
         return noun + "s"
 
 
-def qualified_name(item):
-    # type: (Any) -> str
-    """Attempt to produce the fully qualified name for an item.
-
-    If the item is a type, method, property or function, its fully qualified name is returned as
-    best as can be determined. Otherwise, the fully qualified name of the type of the given item is
-    returned.
-
-    :param item: The item to identify.
-    :return: The fully qualified name of the given item.
-    """
-    if isinstance(item, property):
-        item = item.fget
-    if not hasattr(item, "__name__"):
-        item = type(item)
-    return "{module}.{type}".format(
-        module=getattr(item, "__module__", "<unknown module>"),
-        # There is no __qualname__ in Python 2.7; so we do the best we can.
-        type=getattr(item, "__qualname__", item.__name__),
-    )
-
-
 def safe_copy(source, dest, overwrite=False):
-    # type: (str, str, bool) -> None
+    # type: (Text, Text, bool) -> None
     def do_copy():
         # type: () -> None
         temp_dest = dest + uuid4().hex
         shutil.copy(source, temp_dest)
-        os.rename(temp_dest, dest)
+        safe_rename(temp_dest, dest)
 
     # If the platform supports hard-linking, use that and fall back to copying.
     # Windows does not support hard-linking.
     if hasattr(os, "link"):
         try:
-            os.link(source, dest)
+            safe_link(source, dest)
         except OSError as e:
             if e.errno == errno.EEXIST:
                 # File already exists.  If overwrite=True, write otherwise skip.
@@ -147,7 +144,7 @@ def safe_copy(source, dest, overwrite=False):
                 # `protected_hardlinks` (see: https://www.kernel.org/doc/Documentation/sysctl/fs.txt) and
                 # we can fall back to copying in that case.
                 #
-                # See also https://github.com/pantsbuild/pex/issues/850 where this was discovered.
+                # See also https://github.com/pex-tool/pex/issues/850 where this was discovered.
                 do_copy()
             else:
                 raise
@@ -187,14 +184,36 @@ class MktempTeardownRegistry(object):
 _MKDTEMP_SINGLETON = MktempTeardownRegistry()
 
 
-class PermPreservingZipFile(zipfile.ZipFile, object):
-    """A ZipFile that works around https://bugs.python.org/issue15795."""
+# When Zip64 support was added to Python long ago, the limit was set to
+# half of what it should be. Pex is always dealing with zips it creates
+# via Python and so it does not need to worry about zip implemntations
+# that don't support 4GB; so we monkeypatch our way out of this
+# historical baggage to get working <= 4GB PEXes even on Python 2.7.
+#
+# See: https://github.com/python/cpython/issues/43003
+setattr(zipfile, "ZIP64_LIMIT", (1 << 32) - 1)
+
+
+class ZipFileEx(ZipFile):
+    """A ZipFile that works around several issues in the stdlib.
+
+    See:
+    + https://bugs.python.org/issue15795
+    + https://github.com/pex-tool/pex/issues/298
+    """
 
     class ZipEntry(namedtuple("ZipEntry", ["info", "data"])):
         pass
 
     @classmethod
-    def zip_entry_from_file(cls, filename, arcname=None, date_time=None):
+    def zip_entry_from_file(
+        cls,
+        filename,  # type: str
+        arcname=None,  # type: Optional[str]
+        date_time=None,  # type: Optional[time.struct_time]
+        file_mode=None,  # type: Optional[int]
+    ):
+        # type: (...) -> ZipEntry
         """Construct a ZipEntry for a file on the filesystem.
 
         Usually a similar `zip_info_from_file` method is provided by `ZipInfo`, but it is not
@@ -213,42 +232,142 @@ class PermPreservingZipFile(zipfile.ZipFile, object):
             arcname += "/"
         if date_time is None:
             date_time = time.localtime(st.st_mtime)
-        zinfo = zipfile.ZipInfo(filename=arcname, date_time=date_time[:6])
-        zinfo.external_attr = (st.st_mode & 0xFFFF) << 16  # Unix attributes
+        zip_info = zipfile.ZipInfo(filename=arcname, date_time=date_time[:6])
+
+        # Unix attributes
+        if file_mode:
+            zip_info.external_attr = file_mode << 16
+        else:
+            zip_info.external_attr = (st.st_mode & 0xFFFF) << 16
+
         if isdir:
-            zinfo.file_size = 0
-            zinfo.external_attr |= 0x10  # MS-DOS directory flag
-            zinfo.compress_type = zipfile.ZIP_STORED
+            zip_info.file_size = 0
+            zip_info.external_attr |= 0x10  # MS-DOS directory flag
+            zip_info.compress_type = zipfile.ZIP_STORED
             data = b""
         else:
-            zinfo.file_size = st.st_size
-            zinfo.compress_type = zipfile.ZIP_DEFLATED
+            zip_info.file_size = st.st_size
+            zip_info.compress_type = zipfile.ZIP_DEFLATED
             with open(filename, "rb") as fp:
                 data = fp.read()
-        return cls.ZipEntry(info=zinfo, data=data)
+        return cls.ZipEntry(info=zip_info, data=data)
 
-    def _extract_member(self, member, targetpath, pwd):
-        result = super(PermPreservingZipFile, self)._extract_member(member, targetpath, pwd)
+    def _extract_member(
+        self,
+        member,  # type: Union[str, ZipInfo]
+        targetpath,  # type: str
+        pwd,  # type: Optional[bytes]
+    ):
+        # type: (...) -> str
+
+        # MyPy doesn't see the superclass private method.
+        result = super(ZipFileEx, self)._extract_member(  # type: ignore[misc]
+            member, targetpath, pwd
+        )
         info = member if isinstance(member, zipfile.ZipInfo) else self.getinfo(member)
         self._chmod(info, result)
-        return result
+        return cast(str, result)
+
+    @staticmethod
+    def _chmod(
+        info,  # type: ZipInfo
+        path,  # type: Text
+    ):
+        # type: (...) -> None
 
-    def _chmod(self, info, path):
         # This magic works to extract perm bits from the 32 bit external file attributes field for
         # unix-created zip files, for the layout, see:
         #   https://www.forensicswiki.org/wiki/ZIP#External_file_attributes
-        attr = info.external_attr >> 16
-        os.chmod(path, attr)
+        if info.external_attr > 0xFFFF:
+            attr = info.external_attr >> 16
+            os.chmod(path, attr)
+
+    # Python 3 also takes PathLike[str] for the path arg, but we only ever pass str since we support
+    # Python 2.7 and don't use pathlib as a result.
+    def extractall(  # type: ignore[override]
+        self,
+        path=None,  # type: Optional[str]
+        members=None,  # type: Optional[Iterable[Union[str, ZipInfo]]]
+        pwd=None,  # type: Optional[bytes]
+    ):
+        # type: (...) -> None
+        if sys.version_info[0] != 2:
+            return super(ZipFileEx, self).extractall(path=path, members=members, pwd=pwd)
+
+        # Under Python 2.7, ZipFile does not handle Zip entry name encoding correctly. Older Zip
+        # standards supported IBM code page 437 and newer support UTF-8. The newer support is
+        # indicated by the bit 11 flag in the file header.
+        # From https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT section
+        # "4.4.4 general purpose bit flag: (2 bytes)":
+        #
+        #   Bit 11: Language encoding flag (EFS).  If this bit is set,
+        #           the filename and comment fields for this file
+        #           MUST be encoded using UTF-8. (see APPENDIX D)
+        #
+        # N.B.: MyPy fails to see this code can be reached for Python 2.7.
+        efs_bit = 1 << 11  # type: ignore[unreachable]
+
+        target_path = path or os.getcwd()
+        for member in members or self.infolist():
+            info = member if isinstance(member, ZipInfo) else self.getinfo(member)
+            encoding = "utf-8" if info.flag_bits & efs_bit else "cp437"
+            member_path = info.filename.encode(encoding)
+            target = target_path.encode(encoding)
+
+            rel_dir = os.path.dirname(member_path)
+            abs_dir = os.path.join(target, rel_dir)
+            abs_path = os.path.join(abs_dir, os.path.basename(member_path))
+            if member_path.endswith(b"/"):
+                safe_mkdir(abs_path)
+            else:
+                safe_mkdir(abs_dir)
+                with open(abs_path, "wb") as tfp, self.open(info) as zf_entry:
+                    shutil.copyfileobj(zf_entry, tfp)
+            self._chmod(info, abs_path)
 
 
 @contextlib.contextmanager
-def open_zip(path, *args, **kwargs):
+def open_zip(
+    path,  # type: Text
+    *args,  # type: Any
+    **kwargs  # type: Any
+):
+    # type: (...) -> Iterator[ZipFileEx]
     """A contextmanager for zip files.
 
     Passes through positional and kwargs to zipfile.ZipFile.
     """
-    with contextlib.closing(PermPreservingZipFile(path, *args, **kwargs)) as zip:
-        yield zip
+
+    # allowZip64=True is the default in Python 3.4+ but not in 2.7. We uniformly enable Zip64
+    # extensions across all Pex supported Pythons.
+    kwargs.setdefault("allowZip64", True)
+
+    with contextlib.closing(ZipFileEx(path, *args, **kwargs)) as zip_fp:
+        yield zip_fp
+
+
+def deterministic_walk(*args, **kwargs):
+    # type: (*Any, **Any) -> Iterator[Tuple[str, List[str], List[str]]]
+    """Walk the specified directory tree in deterministic order.
+
+    Takes the same parameters as os.walk and yields tuples of the same shape,
+    except for the `topdown` parameter, which must always be true.
+    `deterministic_walk` is essentially a wrapper of os.walk, and os.walk doesn't
+    allow modifying the order of the walk when called with `topdown` set to false.
+
+    os.walk uses os.listdir or os.scandir, depending on the Python version,
+    both of which don't guarantee the order in which directory entries get listed.
+    So when the build output depends on the order of directory traversal,
+    use deterministic_walk instead.
+    """
+    # when topdown is false, modifying ``dirs`` has no effect
+    assert kwargs.get("topdown", True), "Determinism cannot be guaranteed when ``topdown`` is false"
+    for root, dirs, files in os.walk(*args, **kwargs):
+        dirs.sort()
+        files.sort()
+        yield root, dirs, files
+        # make sure ``dirs`` is sorted after any modifications
+        dirs.sort()
 
 
 @contextlib.contextmanager
@@ -279,7 +398,7 @@ def register_rmtree(directory):
 
 
 def safe_mkdir(directory, clean=False):
-    # type: (str, bool) -> str
+    # type: (_Text, bool) -> _Text
     """Safely create a directory.
 
     Ensures a directory is present.  If it's not there, it is created.  If it is, it's a no-op. If
@@ -292,8 +411,7 @@ def safe_mkdir(directory, clean=False):
     except OSError as e:
         if e.errno != errno.EEXIST:
             raise
-    finally:
-        return directory
+    return directory
 
 
 def safe_open(filename, *args, **kwargs):
@@ -309,7 +427,7 @@ def safe_open(filename, *args, **kwargs):
 
 
 def safe_delete(filename):
-    # type: (str) -> None
+    # type: (Text) -> None
     """Delete a file safely.
 
     If it's not present, no-op.
@@ -322,7 +440,7 @@ def safe_delete(filename):
 
 
 def safe_rmtree(directory):
-    # type: (str) -> None
+    # type: (_Text) -> None
     """Delete a directory if it's present.
 
     If it's not present, no-op.
@@ -348,230 +466,24 @@ def safe_sleep(seconds):
             current_time = time.time()
 
 
-class AtomicDirectory(object):
-    def __init__(self, target_dir):
-        # type: (str) -> None
-        self._target_dir = target_dir
-        self._work_dir = "{}.{}".format(target_dir, uuid4().hex)
-
-    @property
-    def work_dir(self):
-        # type: () -> str
-        return self._work_dir
-
-    @property
-    def target_dir(self):
-        # type: () -> str
-        return self._target_dir
-
-    def is_finalized(self):
-        # type: () -> bool
-        return os.path.exists(self._target_dir)
-
-    def finalize(self, source=None):
-        # type: (Optional[str]) -> None
-        """Rename `work_dir` to `target_dir` using `os.rename()`.
-
-        :param source: An optional source offset into the `work_dir`` to use for the atomic update
-                       of `target_dir`. By default the whole `work_dir` is used.
-
-        If a race is lost and `target_dir` already exists, the `target_dir` dir is left unchanged and
-        the `work_dir` directory will simply be removed.
-        """
-        if self.is_finalized():
-            return
-
-        source = os.path.join(self._work_dir, source) if source else self._work_dir
-        try:
-            # Perform an atomic rename.
-            #
-            # Per the docs: https://docs.python.org/2.7/library/os.html#os.rename
-            #
-            #   The operation may fail on some Unix flavors if src and dst are on different filesystems.
-            #   If successful, the renaming will be an atomic operation (this is a POSIX requirement).
-            #
-            # We have satisfied the single filesystem constraint by arranging the `work_dir` to be a
-            # sibling of the `target_dir`.
-            os.rename(source, self._target_dir)
-        except OSError as e:
-            if e.errno not in (errno.EEXIST, errno.ENOTEMPTY):
-                raise e
-        finally:
-            self.cleanup()
-
-    def cleanup(self):
-        # type: () -> None
-        safe_rmtree(self._work_dir)
-
-
-@contextmanager
-def atomic_directory(target_dir, exclusive, source=None):
-    # type: (str, bool, Optional[str]) -> Iterator[AtomicDirectory]
-    """A context manager that yields a potentially exclusively locked AtomicDirectory.
-
-    :param target_dir: The target directory to atomically update.
-    :param exclusive: If `True`, its guaranteed that only one process will be yielded a non `None`
-                      workdir; otherwise two or more processes might be yielded unique non-`None`
-                      workdirs with the last process to finish "winning".
-    :param source: An optional source offset into the work directory to use for the atomic update
-                   of the target directory. By default the whole work directory is used.
-
-    If the `target_dir` already exists the enclosed block will be yielded an AtomicDirectory that
-    `is_finalized` to signal there is no work to do.
-
-    If the enclosed block fails the `target_dir` will be undisturbed.
-
-    The new work directory will be cleaned up regardless of whether or not the enclosed block
-    succeeds.
-
-    If the contents of the resulting directory will be subsequently mutated it's probably correct to
-    pass `exclusive=True` to ensure mutations that race the creation process are not lost.
-    """
-    atomic_dir = AtomicDirectory(target_dir=target_dir)
-    if atomic_dir.is_finalized():
-        # Our work is already done for us so exit early.
-        yield atomic_dir
-        return
-
-    lock_fd = None  # type: Optional[int]
-
-    def unlock():
-        # type: () -> None
-        if lock_fd is None:
-            return
-        try:
-            fcntl.lockf(lock_fd, fcntl.LOCK_UN)
-        finally:
-            os.close(lock_fd)
-
-    if exclusive:
-        head, tail = os.path.split(atomic_dir.target_dir)
-        if head:
-            safe_mkdir(head)
-        # N.B.: We don't actually write anything to the lock file but the fcntl file locking
-        # operations only work on files opened for at least write.
-        lock_fd = os.open(
-            os.path.join(head, ".{}.atomic_directory.lck".format(tail or "here")),
-            os.O_CREAT | os.O_WRONLY,
-        )
-        # N.B.: Since lockf operates on an open file descriptor and these are guaranteed to be
-        # closed by the operating system when the owning process exits, this lock is immune to
-        # staleness.
-        fcntl.lockf(lock_fd, fcntl.LOCK_EX)  # A blocking write lock.
-        if atomic_dir.is_finalized():
-            # We lost the double-checked locking race and our work was done for us by the race
-            # winner so exit early.
-            try:
-                yield atomic_dir
-            finally:
-                unlock()
-            return
-
-    try:
-        safe_mkdir(atomic_dir.work_dir)
-        yield atomic_dir
-        atomic_dir.finalize(source=source)
-    finally:
-        unlock()
-        atomic_dir.cleanup()
-
-
-def chmod_plus_x(path):
-    # type: (str) -> None
-    """Equivalent of unix `chmod a+x path`"""
-    path_mode = os.stat(path).st_mode
-    path_mode &= int("777", 8)
-    if path_mode & stat.S_IRUSR:
-        path_mode |= stat.S_IXUSR
-    if path_mode & stat.S_IRGRP:
-        path_mode |= stat.S_IXGRP
-    if path_mode & stat.S_IROTH:
-        path_mode |= stat.S_IXOTH
-    os.chmod(path, path_mode)
-
-
-def chmod_plus_w(path):
-    # type: (str) -> None
-    """Equivalent of unix `chmod +w path`"""
-    path_mode = os.stat(path).st_mode
-    path_mode &= int("777", 8)
-    path_mode |= stat.S_IWRITE
-    os.chmod(path, path_mode)
-
-
-def is_exe(path):
-    # type: (str) -> bool
-    """Determines if the given path is a file executable by the current user.
-
-    :param path: The path to check.
-    :return: `True if the given path is a file executable by the current user.
-    """
-    return os.path.isfile(path) and os.access(path, os.R_OK | os.X_OK)
-
-
-def is_script(
-    path,  # type: str
-    pattern=None,  # type: Optional[str]
-    check_executable=True,  # type: bool
+def touch(
+    file,  # type: _Text
+    times=None,  # type: Optional[Union[int, float, Tuple[int, int], Tuple[float, float]]]
 ):
-    # type: (...) -> bool
-    """Determines if the given path is a script.
+    # type: (...) -> _Text
+    """Equivalent of unix `touch path`.
 
-    A script is a file that starts with a shebang (#!...) line.
-
-    :param path: The path to check.
-    :param pattern: An optional pattern to match against the shebang (excluding the leading #!).
-    :param check_executable: Check that the script is executable by the current user.
-    :return: True if the given path is a script.
+    If no times is passed, the current time is used to set atime and mtime. If a single int or float
+    is passed for times, it is used for both atime and mtime. If a 2-tuple of ints or floats is
+    passed, the 1st slot is the atime and the 2nd the mtime, just as for `os.utime`.
     """
-    if check_executable and not is_exe(path):
-        return False
-    with open(path, "rb") as fp:
-        if b"#!" != fp.read(2):
-            return False
-        if not pattern:
-            return True
-        return bool(re.match(pattern, fp.readline().decode("utf-8")))
-
-
-def is_python_script(
-    path,  # type: str
-    check_executable=True,  # type: bool
-):
-    # type: (...) -> bool
-    return is_script(path, pattern=r"(?i)^.*(?:python|pypy)", check_executable=check_executable)
-
-
-def can_write_dir(path):
-    # type: (str) -> bool
-    """Determines if the directory at path can be written to by the current process.
-
-    If the directory doesn't exist, determines if it can be created and thus written to.
-
-    N.B.: This is a best-effort check only that uses permission heuristics and does not actually test
-    that the directory can be written to with and writes.
-
-    :param path: The directory path to test.
-    :return:`True` if the given path is a directory that can be written to by the current process.
-    """
-    while not os.access(path, os.F_OK):
-        parent_path = os.path.dirname(path)
-        if not parent_path or (parent_path == path):
-            # We've recursed up to the root without success, which shouldn't happen,
-            return False
-        path = parent_path
-    return os.path.isdir(path) and os.access(path, os.R_OK | os.W_OK | os.X_OK)
-
-
-def touch(file):
-    # type: (str) -> None
-    """Equivalent of unix `touch path`."""
     with safe_open(file, "a"):
-        os.utime(file, None)
+        os.utime(file, (times, times) if isinstance(times, (int, float)) else times)
+    return file
 
 
 class Chroot(object):
-    """A chroot of files overlayed from one directory to another directory.
+    """A chroot of files overlaid from one directory to another directory.
 
     Files may be tagged when added in order to keep track of multiple overlays in the chroot.
     """
@@ -580,11 +492,7 @@ class Chroot(object):
         pass
 
     class ChrootTaggingException(Error):
-        def __init__(self, filename, orig_tag, new_tag):
-            super(Chroot.ChrootTaggingException, self).__init__(  # noqa: T800
-                "Trying to add %s to fileset(%s) but already in fileset(%s)!"
-                % (filename, new_tag, orig_tag)
-            )
+        pass
 
     def __init__(self, chroot_base):
         # type: (str) -> None
@@ -596,25 +504,10 @@ class Chroot(object):
             safe_mkdir(chroot_base)
         except OSError as e:
             raise self.Error("Unable to create chroot in %s: %s" % (chroot_base, e))
-        self.chroot = chroot_base
-        self.filesets = defaultdict(set)  # type: DefaultDict[str, Set[str]]
-
-    def clone(self, into=None):
-        """Clone this chroot.
-
-        :keyword into: (optional) An optional destination directory to clone the
-          Chroot into.  If not specified, a temporary directory will be created.
-
-        .. versionchanged:: 0.8
-          The temporary directory created when ``into`` is not specified is now garbage collected on
-          interpreter exit.
-        """
-        into = into or safe_mkdtemp()
-        new_chroot = Chroot(into)
-        for label, fileset in self.filesets.items():
-            for fn in fileset:
-                new_chroot.link(os.path.join(self.chroot, fn), fn, label=label)
-        return new_chroot
+        self.chroot = chroot_base  # type: str
+        self.filesets = defaultdict(set)  # type: DefaultDict[Optional[str], Set[str]]
+        self._compress_by_file = {}  # type: Dict[str, bool]
+        self._file_index = {}  # type: Dict[str, Optional[str]]
 
     def path(self):
         # type: () -> str
@@ -622,24 +515,57 @@ class Chroot(object):
         return self.chroot
 
     def _normalize(self, dst):
+        # type: (str) -> str
         dst = os.path.normpath(dst)
         if dst.startswith(os.sep) or dst.startswith(".."):
             raise self.Error("Destination path is not a relative path!")
         return dst
 
-    def _check_tag(self, fn, label):
-        for fs_label, fs in self.filesets.items():
-            if fn in fs and fs_label != label:
-                raise self.ChrootTaggingException(fn, fs_label, label)
+    def _check_tag(
+        self,
+        fn,  # type: str
+        label,  # type: Optional[str]
+        compress=True,  # type: bool
+    ):
+        # type: (...) -> None
+        """Raises ChrootTaggingException if a file was added under more than one label."""
+        existing_label = self._file_index.setdefault(fn, label)
+        if label != existing_label:
+            raise self.ChrootTaggingException(
+                "Trying to add {file} to fileset({new_tag}) but already in "
+                "fileset({orig_tag})!".format(file=fn, new_tag=label, orig_tag=existing_label)
+            )
+        existing_compress = self._compress_by_file.setdefault(fn, compress)
+        if compress != existing_compress:
+            raise self.ChrootTaggingException(
+                "Trying to add {file} to fileset({tag}) with compress {new_compress} but already "
+                "added with compress {orig_compress}!".format(
+                    file=fn, tag=label, new_compress=compress, orig_compress=existing_compress
+                )
+            )
 
-    def _tag(self, fn, label):
-        self._check_tag(fn, label)
+    def _tag(
+        self,
+        fn,  # type: str
+        label,  # type: Optional[str]
+        compress,  # type: bool
+    ):
+        # type: (...) -> None
+        self._check_tag(fn, label, compress)
         self.filesets[label].add(fn)
 
     def _ensure_parent(self, path):
+        # type: (str) -> None
         safe_mkdir(os.path.dirname(os.path.join(self.chroot, path)))
 
-    def copy(self, src, dst, label=None):
+    def copy(
+        self,
+        src,  # type: str
+        dst,  # type: str
+        label=None,  # type: Optional[str]
+        compress=True,  # type: bool
+    ):
+        # type: (...) -> None
         """Copy file ``src`` to ``chroot/dst`` with optional label.
 
         May raise anything shutil.copy can raise, e.g.
@@ -649,11 +575,18 @@ class Chroot(object):
         but with a different label.
         """
         dst = self._normalize(dst)
-        self._tag(dst, label)
+        self._tag(dst, label, compress)
         self._ensure_parent(dst)
         shutil.copy(src, os.path.join(self.chroot, dst))
 
-    def link(self, src, dst, label=None):
+    def link(
+        self,
+        src,  # type: str
+        dst,  # type: str
+        label=None,  # type: Optional[str]
+        compress=True,  # type: bool
+    ):
+        # type: (...) -> None
         """Hard link file from ``src`` to ``chroot/dst`` with optional label.
 
         May raise anything os.link can raise, e.g.
@@ -663,7 +596,7 @@ class Chroot(object):
         but with a different label.
         """
         dst = self._normalize(dst)
-        self._tag(dst, label)
+        self._tag(dst, label, compress)
         self._ensure_parent(dst)
         abs_src = src
         abs_dst = os.path.join(self.chroot, dst)
@@ -675,42 +608,59 @@ class Chroot(object):
         src,  # type: str
         dst,  # type: str
         label=None,  # type: Optional[str]
+        compress=True,  # type: bool
     ):
         # type: (...) -> None
         dst = self._normalize(dst)
-        self._tag(dst, label)
+        self._tag(dst, label, compress)
         self._ensure_parent(dst)
-        abs_src = os.path.abspath(src)
-        abs_dst = os.path.join(self.chroot, dst)
-        os.symlink(abs_src, abs_dst)
+        abs_src = os.path.realpath(src)
+        abs_dst = os.path.realpath(os.path.join(self.chroot, dst))
+        safe_relative_symlink(abs_src, abs_dst)
 
-    def write(self, data, dst, label=None, mode="wb", executable=False):
+    def write(
+        self,
+        data,  # type: Union[str, bytes]
+        dst,  # type: str
+        label=None,  # type: Optional[str]
+        mode="wb",  # type: str
+        executable=False,  # type: bool
+        compress=True,  # type: bool
+    ):
+        # type: (...) -> None
         """Write data to ``chroot/dst`` with optional label.
 
         Has similar exceptional cases as ``Chroot.copy``
         """
         dst = self._normalize(dst)
-        self._tag(dst, label)
+        self._tag(dst, label, compress)
         self._ensure_parent(dst)
         with open(os.path.join(self.chroot, dst), mode) as wp:
             wp.write(data)
         if executable:
             chmod_plus_x(wp.name)
 
-    def touch(self, dst, label=None):
+    def touch(
+        self,
+        dst,  # type: str
+        label=None,  # type: Optional[str]
+    ):
+        # type: (...) -> None
         """Perform 'touch' on ``chroot/dst`` with optional label.
 
         Has similar exceptional cases as Chroot.copy
         """
         dst = self._normalize(dst)
-        self._tag(dst, label)
+        self._tag(dst, label, compress=False)
         touch(os.path.join(self.chroot, dst))
 
     def get(self, label):
+        # type: (Optional[str]) -> Set[str]
         """Get all files labeled with ``label``"""
         return self.filesets.get(label, set())
 
     def files(self):
+        # type: () -> Set[str]
         """Get all files in the chroot."""
         all_files = set()
         for label in self.filesets:
@@ -718,25 +668,29 @@ class Chroot(object):
         return all_files
 
     def labels(self):
+        # type: () -> Iterable[Optional[str]]
         return self.filesets.keys()
 
     def __str__(self):
+        # type: () -> str
         return "Chroot(%s {fs:%s})" % (
             self.chroot,
             " ".join("%s" % foo for foo in self.filesets.keys()),
         )
 
     def delete(self):
+        # type: () -> None
         shutil.rmtree(self.chroot)
 
     def zip(
         self,
         filename,  # type: str
         mode="w",  # type: str
-        deterministic_timestamp=False,  # type: bool
+        deterministic=False,  # type: bool
         exclude_file=lambda _: False,  # type: Callable[[str], bool]
         strip_prefix=None,  # type: Optional[str]
         labels=None,  # type: Optional[Iterable[str]]
+        compress=True,  # type: bool
     ):
         # type: (...) -> None
 
@@ -747,21 +701,29 @@ class Chroot(object):
         else:
             selected_files = self.files()
 
-        with open_zip(filename, mode) as zf:
+        with open_zip(
+            filename, mode, zipfile.ZIP_DEFLATED if compress else zipfile.ZIP_STORED
+        ) as zf:
 
             def write_entry(
                 filename,  # type: str
                 arcname,  # type: str
             ):
                 # type: (...) -> None
+
+                file_mode = None  # type: Optional[int]
+                if deterministic:
+                    file_mode = 0o755 if is_exe(filename) or os.path.isdir(filename) else 0o644
+
                 zip_entry = zf.zip_entry_from_file(
                     filename=filename,
                     arcname=os.path.relpath(arcname, strip_prefix) if strip_prefix else arcname,
-                    date_time=DETERMINISTIC_DATETIME.timetuple()
-                    if deterministic_timestamp
-                    else None,
+                    date_time=(DETERMINISTIC_DATETIME.timetuple() if deterministic else None),
+                    file_mode=file_mode,
                 )
-                zf.writestr(zip_entry.info, zip_entry.data)
+                compress_file = compress and self._compress_by_file.get(arcname, True)
+                compression = zipfile.ZIP_DEFLATED if compress_file else zipfile.ZIP_STORED
+                zf.writestr(zip_entry.info, zip_entry.data, compression)
 
             def get_parent_dir(path):
                 # type: (str) -> Optional[str]
@@ -774,6 +736,8 @@ class Chroot(object):
 
             def maybe_write_parent_dirs(path):
                 # type: (str) -> None
+                if path == strip_prefix:
+                    return
                 parent_dir = get_parent_dir(path)
                 if parent_dir is None or parent_dir in written_dirs:
                     return
@@ -792,8 +756,8 @@ class Chroot(object):
                         yield full_path, path
                         continue
 
-                    for root, _, files in os.walk(full_path):
-                        for f in sorted(files):
+                    for root, _, files in deterministic_walk(full_path):
+                        for f in files:
                             if exclude_file(f):
                                 continue
                             abs_path = os.path.join(root, f)
@@ -803,3 +767,117 @@ class Chroot(object):
             for filename, arcname in iter_files():
                 maybe_write_parent_dirs(arcname)
                 write_entry(filename, arcname)
+
+
+def safe_relative_symlink(
+    src,  # type: Text
+    dst,  # type: Text
+):
+    # type: (...) -> None
+    """Creates a symlink to `src` at `dst` using the relative path to `src` from `dst`.
+
+    :param src: The target of the symlink.
+    :param dst: The path to create the symlink at.
+    """
+    dst_parent = os.path.dirname(dst)
+    safe_mkdir(dst_parent)
+    rel_src = os.path.relpath(src, dst_parent)
+    safe_symlink(rel_src, dst)
+
+
+class CopyMode(Enum["CopyMode.Value"]):
+    class Value(Enum.Value):
+        pass
+
+    COPY = Value("copy")
+    LINK = Value("link")
+    SYMLINK = Value("symlink")
+
+
+CopyMode.seal()
+
+
+def iter_copytree(
+    src,  # type: Text
+    dst,  # type: Text
+    exclude=(),  # type: Container[Text]
+    copy_mode=CopyMode.LINK,  # type: CopyMode.Value
+):
+    # type: (...) -> Iterator[Tuple[Text, Text]]
+    """Copies the directory tree rooted at `src` to `dst` yielding a tuple for each copied file.
+
+    When not using symlinks, if hard links are appropriate they will be used; otherwise files are
+    copied.
+
+    N.B.: The returned iterator must be consumed to drive the copying operations to completion.
+
+    :param src: The source directory tree to copy.
+    :param dst: The destination location to copy the source tree to.
+    :param exclude: Names (basenames) of files and directories to exclude from copying.
+    :param copy_mode: How to copy files.
+    :return: An iterator over tuples identifying the copied files of the form `(src, dst)`.
+    """
+    safe_mkdir(dst)
+    link = copy_mode is CopyMode.LINK
+    for root, dirs, files in os.walk(src, topdown=True, followlinks=True):
+        if src == root:
+            dirs[:] = [d for d in dirs if d not in exclude]
+            files[:] = [f for f in files if f not in exclude]
+
+        for path, is_dir in itertools.chain(
+            zip(dirs, itertools.repeat(True)), zip(files, itertools.repeat(False))
+        ):
+            src_entry = os.path.join(root, path)
+            dst_entry = os.path.join(dst, os.path.relpath(src_entry, src))
+            if not is_dir:
+                yield src_entry, dst_entry
+            try:
+                if copy_mode is CopyMode.SYMLINK:
+                    safe_relative_symlink(src_entry, dst_entry)
+                elif is_dir:
+                    os.mkdir(dst_entry)
+                else:
+                    # We only try to link regular files since linking a symlink on Linux can produce
+                    # another symlink, which leaves open the possibility the src_entry target could
+                    # later go missing leaving the dst_entry dangling.
+                    if link and not os.path.islink(src_entry):
+                        try:
+                            safe_link(src_entry, dst_entry)
+                            continue
+                        except OSError as e:
+                            if e.errno != errno.EXDEV:
+                                raise e
+                            link = False
+                    shutil.copy(src_entry, dst_entry)
+            except OSError as e:
+                if e.errno != errno.EEXIST:
+                    raise e
+
+        if copy_mode is CopyMode.SYMLINK:
+            # Once we've symlinked the top-level directories and files, we've "copied" everything.
+            return
+
+
+@contextmanager
+def environment_as(**kwargs):
+    # type: (**Any) -> Iterator[None]
+    """Mutates the `os.environ` for the duration of the context.
+
+    Keyword arguments with None values are removed from os.environ (if present) and all other
+    keyword arguments are added or updated in `os.environ` with the values taken from the
+    stringification (`str(...)`) of each value.
+    """
+    existing = {key: os.environ.get(key) for key in kwargs}
+
+    def adjust_environment(mapping):
+        for key, value in mapping.items():
+            if value is not None:
+                os.environ[key] = str(value)
+            else:
+                os.environ.pop(key, None)
+
+    adjust_environment(kwargs)
+    try:
+        yield
+    finally:
+        adjust_environment(existing)
