@@ -1,15 +1,13 @@
-# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2021 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
-import json
 import os.path
-import subprocess
 
 from pex import dist_metadata
 from pex.dist_metadata import ProjectNameAndVersion
 from pex.orderedset import OrderedSet
 from pex.pex_info import PexInfo
-from pex.testing import make_env, run_pex_command
 from pex.typing import TYPE_CHECKING
+from testing import run_pex_command, subprocess
 
 if TYPE_CHECKING:
     from typing import Any
@@ -35,5 +33,5 @@ def test_duplicate_requirements_issues_1550(tmpdir):
     assert 1 == len(pex_info.distributions)
     assert ProjectNameAndVersion("PyJWT", "1.7.1") == dist_metadata.project_name_and_version(
         next(iter(pex_info.distributions.keys()))
-    )
+    ), "ex_info.distributions: {}".format(pex_info.distributions)
     assert OrderedSet(("PyJWT", "PyJWT==1.7.1")) == pex_info.requirements
