@@ -1,9 +1,9 @@
-# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2014 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 import os
 import platform
-import subprocess
+import sys
 from contextlib import contextmanager
 from textwrap import dedent
 
@@ -12,6 +12,7 @@ import pytest
 from pex import resolver
 from pex.common import temporary_dir
 from pex.compatibility import to_bytes
+from pex.dist_metadata import Distribution
 from pex.environment import PEXEnvironment, _InvalidWheelName, _RankedDistribution
 from pex.fingerprinted_distribution import FingerprintedDistribution
 from pex.inherit_path import InheritPath
@@ -19,20 +20,22 @@ from pex.interpreter import PythonInterpreter
 from pex.pex import PEX
 from pex.pex_builder import PEXBuilder
 from pex.pex_info import PexInfo
-from pex.rank import Rank
+from pex.resolve.configured_resolver import ConfiguredResolver
 from pex.targets import LocalInterpreter, Targets
-from pex.testing import (
-    IS_LINUX,
+from pex.typing import TYPE_CHECKING
+from testing import (
+    IS_LINUX_X86_64,
     IS_PYPY3,
-    PY37,
+    PY39,
     WheelBuilder,
     ensure_python_interpreter,
+    install_wheel,
     make_bdist,
+    subprocess,
     temporary_content,
     temporary_filename,
 )
-from pex.third_party.pkg_resources import Distribution
-from pex.typing import TYPE_CHECKING
+from testing.dist_metadata import create_dist_metadata
 
 if TYPE_CHECKING:
     from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Tuple
@@ -52,7 +55,13 @@ def assert_force_local_implicit_ns_packages_issues_598(
 ):
     def create_foo_bar_setup(name, **extra_args):
         # type: (str, **Any) -> str
-        setup_args = dict(name=name, version="0.0.1", packages=["foo", "foo.bar"])
+        setup_args = dict(
+            name=name,
+            version="0.0.1",
+            packages=["foo", "foo.bar"],
+            # Ensure we generate a universal wheel:
+            options={"bdist_wheel": {"python_tag": "py2.py3"}},
+        )
         if create_ns_packages:
             setup_args.update(namespace_packages=["foo", "foo.bar"])
         if requirements:
@@ -131,22 +140,22 @@ def assert_force_local_implicit_ns_packages_issues_598(
         }
     )
 
-    def add_requirements(builder, cache):
-        # type: (PEXBuilder, str) -> None
-        for installed_dist in resolver.resolve(
+    def add_requirements(builder):
+        # type: (PEXBuilder) -> None
+        for resolved_dist in resolver.resolve(
             targets=Targets(interpreters=(builder.interpreter,)),
             requirements=requirements,
-            cache=cache,
-        ).installed_distributions:
-            builder.add_distribution(installed_dist.distribution)
-            for direct_req in installed_dist.direct_requirements:
+            resolver=ConfiguredResolver.default(),
+        ).distributions:
+            builder.add_distribution(resolved_dist.distribution)
+            for direct_req in resolved_dist.direct_requirements:
                 builder.add_requirement(direct_req)
 
     def add_wheel(builder, content):
         # type: (PEXBuilder, Dict[str, str]) -> None
         with temporary_content(content) as project:
-            dist = WheelBuilder(project, interpreter=builder.interpreter).bdist()
-            builder.add_dist_location(dist)
+            dist = install_wheel(WheelBuilder(project).bdist())
+            builder.add_dist_location(dist.location)
 
     def add_sources(builder, content):
         # type: (PEXBuilder, Dict[str, str]) -> None
@@ -154,20 +163,24 @@ def assert_force_local_implicit_ns_packages_issues_598(
             for path in content.keys():
                 builder.add_source(os.path.join(project, path), path)
 
-    with temporary_dir() as root, temporary_dir() as cache:
+    with temporary_dir() as root:
+        pex_root = os.path.join(root, "pex_root")
+
         pex_info1 = PexInfo.default()
+        pex_info1.pex_root = pex_root
         pex1 = os.path.join(root, "pex1.pex")
         builder1 = PEXBuilder(interpreter=interpreter, pex_info=pex_info1)
-        add_requirements(builder1, cache)
+        add_requirements(builder1)
         add_wheel(builder1, content1)
         add_sources(builder1, content2)
         builder1.build(pex1)
 
         pex_info2 = PexInfo.default()
-        pex_info2.pex_path = pex1
+        pex_info2.pex_root = pex_root
+        pex_info2.pex_path = [pex1]
         pex2 = os.path.join(root, "pex2")
         builder2 = PEXBuilder(path=pex2, interpreter=interpreter, pex_info=pex_info2)
-        add_requirements(builder2, cache)
+        add_requirements(builder2)
         add_wheel(builder2, content3)
         builder2.set_script("foobaz")
         builder2.freeze()
@@ -187,7 +200,19 @@ def get_setuptools_requirement(interpreter=None):
     )
 
 
-@pytest.mark.xfail(IS_PYPY3, reason="https://github.com/pantsbuild/pex/issues/1210")
+skip_if_no_pkg_resources = pytest.mark.skipif(
+    sys.version_info[:2] >= (3, 12),
+    reason=(
+        "These tests requires pex.third_party.pkg_resources from vendored setuptools and that "
+        "version of pkg_resources uses a vendor meta path importer that only implements the "
+        "PEP-302 finder spec and not the modern spec. Only the modern finder spec is supported by "
+        "Python 3.12+."
+    ),
+)
+
+
+@skip_if_no_pkg_resources
+@pytest.mark.xfail(IS_PYPY3, reason="https://github.com/pex-tool/pex/issues/1210")
 def test_issues_598_explicit_any_interpreter():
     # type: () -> None
     assert_force_local_implicit_ns_packages_issues_598(
@@ -195,30 +220,33 @@ def test_issues_598_explicit_any_interpreter():
     )
 
 
+@skip_if_no_pkg_resources
 def test_issues_598_explicit_missing_requirement():
     # type: () -> None
     assert_force_local_implicit_ns_packages_issues_598(create_ns_packages=True)
 
 
 @pytest.fixture
-def python_37_interpreter():
+def python_39_interpreter():
     # type: () -> PythonInterpreter
     # Python 3.7 supports implicit namespace packages.
-    return PythonInterpreter.from_binary(ensure_python_interpreter(PY37))
+    return PythonInterpreter.from_binary(ensure_python_interpreter(PY39))
 
 
-def test_issues_598_implicit(python_37_interpreter):
+@skip_if_no_pkg_resources
+def test_issues_598_implicit(python_39_interpreter):
     # type: (PythonInterpreter) -> None
     assert_force_local_implicit_ns_packages_issues_598(
-        interpreter=python_37_interpreter, create_ns_packages=False
+        interpreter=python_39_interpreter, create_ns_packages=False
     )
 
 
-def test_issues_598_implicit_explicit_mixed(python_37_interpreter):
+@skip_if_no_pkg_resources
+def test_issues_598_implicit_explicit_mixed(python_39_interpreter):
     # type: (PythonInterpreter) -> None
     assert_force_local_implicit_ns_packages_issues_598(
-        interpreter=python_37_interpreter,
-        requirements=[get_setuptools_requirement(python_37_interpreter)],
+        interpreter=python_39_interpreter,
+        requirements=[get_setuptools_requirement(python_39_interpreter)],
         create_ns_packages=True,
     )
 
@@ -247,78 +275,86 @@ def test_osx_platform_intel_issue_523():
         # type: () -> PythonInterpreter
         return PythonInterpreter.from_binary(_KNOWN_BAD_APPLE_INTERPRETER)
 
-    with temporary_dir() as cache:
-        # We need to run the bad interpreter with a modern, non-Apple-Extras setuptools in order to
-        # successfully install psutil; yield_pex_builder sets up the bad interpreter with our vendored
-        # setuptools and wheel extras.
-        with yield_pex_builder(
-            interpreter=bad_interpreter()
-        ) as pb, temporary_filename() as pex_file:
-            for installed_dist in resolver.resolve(
-                targets=Targets(interpreters=(pb.interpreter,)),
-                requirements=["psutil==5.4.3"],
-                cache=cache,
-            ).installed_distributions:
-                pb.add_dist_location(installed_dist.distribution.location)
-            pb.build(pex_file)
+    # We need to run the bad interpreter with a modern, non-Apple-Extras setuptools in order to
+    # successfully install psutil; yield_pex_builder sets up the bad interpreter with our vendored
+    # setuptools and wheel extras.
+    with yield_pex_builder(interpreter=bad_interpreter()) as pb, temporary_filename() as pex_file:
+        for resolved_dist in resolver.resolve(
+            targets=Targets(interpreters=(pb.interpreter,)),
+            requirements=["psutil==5.4.3"],
+            resolver=ConfiguredResolver.default(),
+        ).distributions:
+            pb.add_dist_location(resolved_dist.distribution.location)
+        pb.build(pex_file)
 
-            # NB: We want PEX to find the bare bad interpreter at runtime.
-            pex = PEX(pex_file, interpreter=bad_interpreter())
+        # NB: We want PEX to find the bare bad interpreter at runtime.
+        pex = PEX(pex_file, interpreter=bad_interpreter())
 
-            def run(args, **env):
-                # type: (Iterable[str], **str) -> Tuple[int, str, str]
-                pex_env = os.environ.copy()
-                pex_env["PEX_VERBOSE"] = "1"
-                pex_env.update(**env)
-                process = pex.run(
-                    args=args,
-                    env=pex_env,
-                    blocking=False,
-                    stdout=subprocess.PIPE,
-                    stderr=subprocess.PIPE,
-                )
-                stdout, stderr = process.communicate()
-                return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")
-
-            returncode, _, stderr = run(["-c", "import psutil"])
-            assert 0 == returncode, "Process failed with exit code {} and stderr:\n{}".format(
-                returncode, stderr
+        def run(args, **env):
+            # type: (Iterable[str], **str) -> Tuple[int, str, str]
+            pex_env = os.environ.copy()
+            pex_env["PEX_VERBOSE"] = "1"
+            pex_env.update(**env)
+            process = pex.run(
+                args=args,
+                env=pex_env,
+                blocking=False,
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
             )
+            stdout, stderr = process.communicate()
+            return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")
 
-            returncode, stdout, stderr = run(["-c", "import pkg_resources"])
-            assert 0 != returncode, (
-                "Isolated pex process succeeded but should not have found pkg-resources:\n"
-                "STDOUT:\n"
-                "{}\n"
-                "STDERR:\n"
-                "{}".format(stdout, stderr)
-            )
+        returncode, _, stderr = run(["-c", "import psutil"])
+        assert 0 == returncode, "Process failed with exit code {} and stderr:\n{}".format(
+            returncode, stderr
+        )
 
-            returncode, stdout, stderr = run(
-                ["-c", "import pkg_resources; print(pkg_resources.get_supported_platform())"],
-                # Let the bad interpreter site-packages setuptools leak in.
-                PEX_INHERIT_PATH=InheritPath.for_value(True).value,
-            )
-            assert 0 == returncode, "Process failed with exit code {} and stderr:\n{}".format(
-                returncode, stderr
-            )
+        returncode, stdout, stderr = run(["-c", "import pkg_resources"])
+        assert 0 != returncode, (
+            "Isolated pex process succeeded but should not have found pkg-resources:\n"
+            "STDOUT:\n"
+            "{}\n"
+            "STDERR:\n"
+            "{}".format(stdout, stderr)
+        )
 
-            # Verify this worked along side the previously problematic pkg_resources-reported platform.
-            release, _, _ = platform.mac_ver()
-            major_minor = ".".join(release.split(".")[:2])
-            assert "macosx-{}-intel".format(major_minor) == stdout.strip()
+        returncode, stdout, stderr = run(
+            ["-c", "import pkg_resources; print(pkg_resources.get_supported_platform())"],
+            # Let the bad interpreter site-packages setuptools leak in.
+            PEX_INHERIT_PATH=InheritPath.for_value(True).value,
+        )
+        assert 0 == returncode, "Process failed with exit code {} and stderr:\n{}".format(
+            returncode, stderr
+        )
+
+        # Verify this worked along side the previously problematic pkg_resources-reported platform.
+        release, _, _ = platform.mac_ver()
+        major_minor = ".".join(release.split(".")[:2])
+        machine = platform.machine()
+        assert "macosx-{}-{}".format(major_minor, machine) == stdout.strip()
 
 
+@pytest.mark.skipif(
+    sys.version_info[:2] >= (3, 12),
+    reason=(
+        "Pex 1.6.3 attempts an import of pex.third_party.pkg_resources from vendored setuptools "
+        "and that version of pkg_resources uses a vendor meta path importer that only implements "
+        "the PEP-302 finder spec and not the modern spec. Only the modern finder spec is supported "
+        "by Python 3.12+."
+    ),
+)
 def test_activate_extras_issue_615():
     # type: () -> None
     with yield_pex_builder() as pb:
-        for installed_dist in resolver.resolve(
+        for resolved_dist in resolver.resolve(
             targets=Targets(interpreters=(pb.interpreter,)),
             requirements=["pex[requests]==1.6.3"],
-        ).installed_distributions:
-            for direct_req in installed_dist.direct_requirements:
+            resolver=ConfiguredResolver.default(),
+        ).distributions:
+            for direct_req in resolved_dist.direct_requirements:
                 pb.add_requirement(direct_req)
-            pb.add_dist_location(installed_dist.distribution.location)
+            pb.add_dist_location(resolved_dist.distribution.location)
         pb.set_script("pex")
         pb.freeze()
         process = PEX(pb.path(), interpreter=pb.interpreter).run(
@@ -330,7 +366,7 @@ def test_activate_extras_issue_615():
         )
         stdout, stderr = process.communicate()
         assert 0 == process.returncode, "Process failed with exit code {} and output:\n{}".format(
-            process.returncode, stderr
+            process.returncode, stderr.decode("utf-8")
         )
         assert to_bytes("{} 1.6.3".format(os.path.basename(pb.path()))) == stdout.strip()
 
@@ -339,23 +375,31 @@ def assert_namespace_packages_warning(distribution, version, expected_warning):
     # type: (str, str, bool) -> None
     requirement = "{}=={}".format(distribution, version)
     pb = PEXBuilder()
-    for installed_dist in resolver.resolve(requirements=[requirement]).installed_distributions:
-        pb.add_dist_location(installed_dist.distribution.location)
+    for resolved_dist in resolver.resolve(
+        requirements=[requirement],
+        resolver=ConfiguredResolver.default(),
+    ).distributions:
+        pb.add_dist_location(resolved_dist.distribution.location)
     pb.freeze()
 
     process = PEX(pb.path()).run(args=["-c", ""], blocking=False, stderr=subprocess.PIPE)
     _, stderr = process.communicate()
     stderr_text = stderr.decode("utf8")
 
-    partial_warning_preamble = "PEXWarning: The `pkg_resources` package was loaded"
+    if sys.version_info[:2] >= (3, 12):
+        partial_warning_preamble = (
+            "PEXWarning: The legacy `pkg_resources` package cannot be imported by"
+        )
+    else:
+        partial_warning_preamble = "PEXWarning: The `pkg_resources` package was loaded"
     partial_warning_detail = "{} namespace packages:".format(requirement)
 
     if expected_warning:
-        assert partial_warning_preamble in stderr_text
-        assert partial_warning_detail in stderr_text
+        assert partial_warning_preamble in stderr_text, stderr_text
+        assert partial_warning_detail in stderr_text, stderr_text
     else:
-        assert partial_warning_preamble not in stderr_text
-        assert partial_warning_detail not in stderr_text
+        assert partial_warning_preamble not in stderr_text, stderr_text
+        assert partial_warning_detail not in stderr_text, stderr_text
 
 
 def test_present_non_empty_namespace_packages_metadata_does_warn():
@@ -370,89 +414,98 @@ def test_present_but_empty_namespace_packages_metadata_does_not_warn():
 
 def create_dist(
     location,  # type: str
-    version="1.0.0",  # type: Optional[str]
+    project_name="foo",  # type: str
+    version="1.0.0",  # type: str
 ):
     # type: (...) -> FingerprintedDistribution
-    # N.B.: version must be set simply so that __hash__ / __eq__ work correctly in the
-    # `pex.dist_metadata` module.
     return FingerprintedDistribution(
-        distribution=Distribution(location=location, version=version), fingerprint=location
+        distribution=Distribution(
+            location=location,
+            metadata=create_dist_metadata(
+                project_name=project_name, version=version, location=location
+            ),
+        ),
+        fingerprint=location,
     )
 
 
 @pytest.fixture
-def cpython_37_environment(python_37_interpreter):
+def cpython_39_environment(python_39_interpreter):
     return PEXEnvironment(
         pex="",
-        pex_info=PexInfo.default(python_37_interpreter),
-        target=LocalInterpreter.create(python_37_interpreter),
+        pex_info=PexInfo.default(),
+        target=LocalInterpreter.create(python_39_interpreter),
     )
 
 
 @pytest.mark.parametrize(
-    ("wheel_distribution", "wheel_is_linux"),
+    ("wheel_distribution", "wheel_is_linux_x86_64"),
     [
         pytest.param(
-            create_dist("llvmlite-0.29.0-cp37-cp37m-linux_x86_64.whl", "0.29.0"),
+            create_dist("llvmlite-0.29.0-cp39-cp39-linux_x86_64.whl", "llvmlite", "0.29.0"),
             True,
             id="without_build_tag_linux",
         ),
         pytest.param(
-            create_dist("llvmlite-0.29.0-1-cp37-cp37m-linux_x86_64.whl", "0.29.0"),
+            create_dist("llvmlite-0.29.0-1-cp39-cp39-linux_x86_64.whl", "llvmlite", "0.29.0"),
             True,
             id="with_build_tag_linux",
         ),
         pytest.param(
-            create_dist("llvmlite-0.29.0-cp37-cp37m-macosx_10.9_x86_64.whl", "0.29.0"),
+            create_dist("llvmlite-0.29.0-cp39-cp39-macosx_10.9_x86_64.whl", "llvmlite", "0.29.0"),
             False,
             id="without_build_tag_osx",
         ),
         pytest.param(
-            create_dist("llvmlite-0.29.0-1-cp37-cp37m-macosx_10.9_x86_64.whl", "0.29.0"),
+            create_dist("llvmlite-0.29.0-1-cp39-cp39-macosx_10.9_x86_64.whl", "llvmlite", "0.29.0"),
             False,
             id="with_build_tag_osx",
         ),
     ],
 )
 def test_can_add_handles_optional_build_tag_in_wheel(
-    cpython_37_environment, wheel_distribution, wheel_is_linux
+    cpython_39_environment, wheel_distribution, wheel_is_linux_x86_64
 ):
     # type: (PEXEnvironment, FingerprintedDistribution, bool) -> None
-    native_wheel = IS_LINUX and wheel_is_linux
-    added = isinstance(cpython_37_environment._can_add(wheel_distribution), _RankedDistribution)
+    native_wheel = IS_LINUX_X86_64 and wheel_is_linux_x86_64
+    added = isinstance(cpython_39_environment._can_add(wheel_distribution), _RankedDistribution)
     assert added is native_wheel
 
 
-def test_can_add_handles_invalid_wheel_filename(cpython_37_environment):
+def test_can_add_handles_invalid_wheel_filename(cpython_39_environment):
     # type: (PEXEnvironment) -> None
     dist = create_dist("pep427-invalid.whl")
-    assert _InvalidWheelName(dist, "pep427-invalid") == cpython_37_environment._can_add(dist)
+    assert _InvalidWheelName(dist, "pep427-invalid.whl") == cpython_39_environment._can_add(dist)
 
 
 @pytest.fixture
-def assert_cpython_37_environment_can_add(cpython_37_environment):
+def assert_cpython_39_environment_can_add(cpython_39_environment):
     # type: (PEXEnvironment) -> Callable[[FingerprintedDistribution], _RankedDistribution]
-    def assert_can_add(dist):
-        # type: (Distribution) -> _RankedDistribution
-        rank = cpython_37_environment._can_add(dist)
+    def assert_can_add(fingerprinted_dist):
+        # type: (FingerprintedDistribution) -> _RankedDistribution
+        rank = cpython_39_environment._can_add(fingerprinted_dist)
         assert isinstance(rank, _RankedDistribution)
         return rank
 
     return assert_can_add
 
 
-def test_can_add_ranking_platform_tag_more_specific(assert_cpython_37_environment_can_add):
-    # type: (Callable[[Distribution], _RankedDistribution]) -> None
-    ranked_specific = assert_cpython_37_environment_can_add(
-        create_dist("foo-1.0.0-cp37-cp37m-macosx_10_9_x86_64.linux_x86_64.whl", "1.0.0")
+# So the following tests pass on all platforms.
+_wheel_tags = "macosx_10_9_x86_64.macosx_11_0_arm64.linux_x86_64.linux_aarch64"
+
+
+def test_can_add_ranking_platform_tag_more_specific(assert_cpython_39_environment_can_add):
+    # type: (Callable[[FingerprintedDistribution], _RankedDistribution]) -> None
+    ranked_specific = assert_cpython_39_environment_can_add(
+        create_dist("foo-1.0.0-cp39-cp39-{}.whl".format(_wheel_tags), "foo", "1.0.0")
     )
-    ranked_universal = assert_cpython_37_environment_can_add(
-        create_dist("foo-2.0.0-py2.py3-none-any.whl", "2.0.0")
+    ranked_universal = assert_cpython_39_environment_can_add(
+        create_dist("foo-2.0.0-py2.py3-none-any.whl", "foo", "2.0.0")
     )
     assert ranked_specific < ranked_universal
 
-    ranked_almost_py3universal = assert_cpython_37_environment_can_add(
-        create_dist("foo-2.0.0-py3-none-any.whl", "2.0.0")
+    ranked_almost_py3universal = assert_cpython_39_environment_can_add(
+        create_dist("foo-2.0.0-py3-none-any.whl", "foo", "2.0.0")
     )
     assert ranked_universal.rank == ranked_almost_py3universal.rank, (
         "Expected the 'universal' compressed tag set to be expanded into two tags and the more "
@@ -460,12 +513,12 @@ def test_can_add_ranking_platform_tag_more_specific(assert_cpython_37_environmen
     )
 
 
-def test_can_add_ranking_version_newer_tie_break(assert_cpython_37_environment_can_add):
-    # type: (Callable[[Distribution], _RankedDistribution]) -> None
-    ranked_v1 = assert_cpython_37_environment_can_add(
-        create_dist("foo-1.0.0-cp37-cp37m-macosx_10_9_x86_64.linux_x86_64.whl", "1.0.0")
+def test_can_add_ranking_version_newer_tie_break(assert_cpython_39_environment_can_add):
+    # type: (Callable[[FingerprintedDistribution], _RankedDistribution]) -> None
+    ranked_v1 = assert_cpython_39_environment_can_add(
+        create_dist("foo-1.0.0-cp39-cp39-{}.whl".format(_wheel_tags), "foo", "1.0.0")
     )
-    ranked_v2 = assert_cpython_37_environment_can_add(
-        create_dist("foo-2.0.0-cp37-cp37m-macosx_10_9_x86_64.linux_x86_64.whl", "2.0.0")
+    ranked_v2 = assert_cpython_39_environment_can_add(
+        create_dist("foo-2.0.0-cp39-cp39-{}.whl".format(_wheel_tags), "foo", "2.0.0")
     )
     assert ranked_v2 < ranked_v1
