@@ -1,4 +1,4 @@
-# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2020 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 from __future__ import absolute_import
@@ -8,10 +8,12 @@ from textwrap import dedent
 
 import pytest
 
-from pex.common import safe_open, temporary_dir, touch
+from pex.artifact_url import VCS, ArtifactURL
+from pex.common import environment_as, safe_open, touch
+from pex.compatibility import urlparse
+from pex.dist_metadata import Requirement
 from pex.fetcher import URLFetcher
 from pex.requirements import (
-    VCS,
     Constraint,
     LocalProjectRequirement,
     LogicalLine,
@@ -20,18 +22,16 @@ from pex.requirements import (
     Source,
     URLRequirement,
     VCSRequirement,
-    VCSScheme,
     parse_requirement_file,
     parse_requirement_from_project_name_and_specifier,
     parse_requirements,
 )
-from pex.testing import environment_as
 from pex.third_party.packaging.markers import Marker
-from pex.third_party.pkg_resources import Requirement
 from pex.typing import TYPE_CHECKING
+from testing.pytest_utils.tmp import Tempdir
 
 if TYPE_CHECKING:
-    from typing import Any, Iterable, Iterator, List, Optional, Union
+    from typing import Any, Iterable, List, Optional, Union
 
     import attr  # vendor:skip
 
@@ -42,19 +42,11 @@ else:
     from pex.third_party import attr
 
 
-@pytest.fixture
-def chroot():
-    # type: () -> Iterator[str]
-    with temporary_dir() as chroot:
-        curdir = os.getcwd()
-        try:
-            os.chdir(chroot)
-            yield chroot
-        finally:
-            os.chdir(curdir)
+def test_parse_requirements_failure_bad_include(tmpdir):
+    # type: (Tempdir) -> None
 
-
-def test_parse_requirements_failure_bad_include(chroot):
+    # N.B.: This test assumes there is no `other-requirements.txt` in the root of the Pex repo,
+    # which is the cwd of the test runner.
     req_iter = parse_requirements(Source.from_text("\n-r other-requirements.txt"))
     with pytest.raises(ParseError) as exc_info:
         next(req_iter)
@@ -68,9 +60,9 @@ def test_parse_requirements_failure_bad_include(chroot):
     )
 
 
-def test_parse_requirements_failure_bad_requirement(chroot):
-    # type: (str) -> None
-    other_requirement_file = os.path.realpath(os.path.join(chroot, "other-requirements.txt"))
+def test_parse_requirements_failure_bad_requirement(tmpdir):
+    # type: (Tempdir) -> None
+    other_requirement_file = tmpdir.join("other-requirements.txt")
     with safe_open(other_requirement_file, "w") as fp:
         fp.write(
             dedent(
@@ -90,15 +82,17 @@ def test_parse_requirements_failure_bad_requirement(chroot):
             )
         )
 
-    req_iter = parse_requirements(Source.from_text("-r other-requirements.txt"))
+    req_iter = parse_requirements(
+        Source.from_text("-r {requirements_txt}".format(requirements_txt=other_requirement_file))
+    )
 
     parsed_requirement = next(req_iter)
     assert isinstance(parsed_requirement, PyPIRequirement)
-    assert "GoodRequirement" == parsed_requirement.requirement.project_name
+    assert "GoodRequirement" == parsed_requirement.requirement.name
 
     parsed_requirement = next(req_iter)
     assert isinstance(parsed_requirement, PyPIRequirement)
-    assert "AnotherRequirement" == parsed_requirement.requirement.project_name
+    assert "AnotherRequirement" == parsed_requirement.requirement.name
 
     with pytest.raises(ParseError) as exc_info:
         next(req_iter)
@@ -135,7 +129,6 @@ def req(
     extras=None,  # type: Optional[Iterable[str]]
     specifier=None,  # type: Optional[str]
     marker=None,  # type: Optional[str]
-    editable=False,  # type: bool
 ):
     # type: (...) -> PyPIRequirement
     return PyPIRequirement(
@@ -143,7 +136,23 @@ def req(
         requirement=parse_requirement_from_project_name_and_specifier(
             project_name, extras=extras, specifier=specifier, marker=marker
         ),
-        editable=editable,
+    )
+
+
+def file_req(
+    url,  # type: str
+    project_name,  # type: str
+    extras=None,  # type: Optional[Iterable[str]]
+    specifier=None,  # type: Optional[str]
+    marker=None,  # type: Optional[str]
+):
+    # type: (...) -> URLRequirement
+    return URLRequirement(
+        line=DUMMY_LINE,
+        url=ArtifactURL.parse(url),
+        requirement=parse_requirement_from_project_name_and_specifier(
+            project_name, extras=extras, specifier=specifier, marker=marker
+        ),
     )
 
 
@@ -153,16 +162,14 @@ def url_req(
     extras=None,  # type: Optional[Iterable[str]]
     specifier=None,  # type: Optional[str]
     marker=None,  # type: Optional[str]
-    editable=False,  # type: bool
 ):
     # type: (...) -> URLRequirement
     return URLRequirement(
         line=DUMMY_LINE,
-        url=url,
+        url=ArtifactURL.parse(url),
         requirement=parse_requirement_from_project_name_and_specifier(
-            project_name, extras=extras, specifier=specifier, marker=marker
+            project_name, extras=extras, specifier=specifier, marker=marker, url=url
         ),
-        editable=editable,
     )
 
 
@@ -173,17 +180,23 @@ def vcs_req(
     extras=None,  # type: Optional[Iterable[str]]
     specifier=None,  # type: Optional[str]
     marker=None,  # type: Optional[str]
-    editable=False,  # type: bool
 ):
     # type: (...) -> VCSRequirement
+
+    url_info = urlparse.urlparse(url)
     return VCSRequirement(
         line=DUMMY_LINE,
         vcs=vcs,
         url=url,
         requirement=parse_requirement_from_project_name_and_specifier(
-            project_name, extras=extras, specifier=specifier, marker=marker
+            project_name,
+            extras=extras,
+            specifier=specifier,
+            marker=marker,
+            url=url_info._replace(
+                scheme="{vcs}+{scheme}".format(vcs=vcs, scheme=url_info.scheme)
+            ).geturl(),
         ),
-        editable=editable,
     )
 
 
@@ -215,9 +228,10 @@ def normalize_results(parsed_requirements):
     ]
 
 
-def test_parse_requirements_stress(chroot):
-    # type: (str) -> None
-    with safe_open(os.path.join(chroot, "other-requirements.txt"), "w") as fp:
+def test_parse_requirements_stress(tmpdir):
+    # type: (Tempdir) -> None
+    other_requirements_txt = tmpdir.join("other-requirements.txt")
+    with safe_open(other_requirements_txt, "w") as fp:
         fp.write(
             # This includes both example snippets taken directly from
             # https://pip.pypa.io/en/stable/reference/pip_install/#requirements-file-format
@@ -251,9 +265,9 @@ def test_parse_requirements_stress(chroot):
                 """
             )
         )
-    touch("somewhere/over/here/pyproject.toml")
+    touch(tmpdir.join("somewhere", "over", "here", "pyproject.toml"))
 
-    with safe_open(os.path.join(chroot, "extra", "stress.txt"), "w") as fp:
+    with safe_open(tmpdir.join("extra", "stress.txt"), "w") as fp:
         fp.write(
             # These are tests of edge cases not included anywhere in the examples found in
             # https://pip.pypa.io/en/stable/reference/pip_install/#requirements-file-format.
@@ -263,6 +277,7 @@ def test_parse_requirements_stress(chroot):
 
                 a/local/project[foo]; python_full_version == "2.7.8"
                 ./another/local/project;python_version == "2.7.*"
+                ./another/local/project ; python_version == "2.7.*"
                 ./another/local/project
                 ./
                 # Local projects with basenames that are invalid Python project names (trailing _):
@@ -272,28 +287,42 @@ def test_parse_requirements_stress(chroot):
 
                 hg+http://hg.example.com/MyProject@da39a3ee5e6b\\
                     #egg=AnotherProject[extra,more];python_version=="3.9.*"&subdirectory=foo/bar
+                hg+http://hg.example.com/MyProject@da39a3ee5e6b\\
+                    #egg=AnotherProject[extra,more] ; python_version=="3.9.*"&subdirectory=foo/bar
 
-                ftp://a/${PROJECT_NAME}-1.0.tar.gz
-                http://a/${PROJECT_NAME}-1.0.zip
+                ftp://a/${{PROJECT_NAME}}-1.0.tar.gz
+                http://a/${{PROJECT_NAME}}-1.0.zip
                 https://a/numpy-1.9.2-cp34-none-win32.whl
                 https://a/numpy-1.9.2-cp34-none-win32.whl;\\
                     python_version=="3.4.*" and sys_platform=='win32'
+                https://a/numpy-1.9.2-cp34-none-win32.whl ; \\
+                    python_version=="3.4.*" and sys_platform=='win32'
 
                 Django@ git+https://github.com/django/django.git
                 Django@git+https://github.com/django/django.git@stable/2.1.x
                 Django@ git+https://github.com/django/django.git\\
                     @fd209f62f1d83233cc634443cfac5ee4328d98b8
                 Django @ file:projects/django-2.3.zip; python_version >= "3.10"
-                """
-            )
-        )
-    touch("extra/pyproject.toml")
-    touch("extra/a/local/project/pyproject.toml")
-    touch("extra/another/local/project/setup.py")
-    touch("extra/tmp/tmpW8tdb_/setup.py")
-    touch("extra/projects/django-2.3.zip")
+                Django @ file:projects/django-2.3.zip ;python_version >= "3.10"
 
-    with safe_open(os.path.join(chroot, "subdir", "more-requirements.txt"), "w") as fp:
+                # Wheel with local version
+                http://download.pytorch.org/whl/cpu/torch-1.12.1%2Bcpu-cp310-cp310-linux_x86_64.whl
+
+                # Editable
+                -e file://{chroot}/extra/a/local/project
+                --editable file://{chroot}/extra/a/local/project/
+                -e ./another/local/project
+                --editable ./another/local/project/
+                """
+            ).format(chroot=tmpdir)
+        )
+    touch(tmpdir.join("extra", "pyproject.toml"))
+    touch(tmpdir.join("extra", "a", "local", "project", "pyproject.toml"))
+    touch(tmpdir.join("extra", "another", "local", "project", "setup.py"))
+    touch(tmpdir.join("extra", "tmp", "tmpW8tdb_", "setup.py"))
+    touch(tmpdir.join("extra", "projects", "django-2.3.zip"))
+
+    with safe_open(tmpdir.join("extra", "subdir", "more-requirements.txt"), "w") as fp:
         fp.write(
             # This checks requirements (`ReqInfo`s) are wrapped up into `Constraints`.
             dedent(
@@ -303,8 +332,9 @@ def test_parse_requirements_stress(chroot):
             )
         )
 
-    req_iter = parse_requirements(
-        Source.from_text(
+    root_requirements_txt = tmpdir.join("requirements.txt")
+    with safe_open(root_requirements_txt, "w") as fp:
+        fp.write(
             # N.B.: Taken verbatim from:
             #   https://pip.pypa.io/en/stable/reference/pip_install/#example-requirements-file
             dedent(
@@ -318,7 +348,7 @@ def test_parse_requirements_stress(chroot):
                 beautifulsoup4
                 #
                 ###### Requirements with Version Specifiers ######
-                #   See https://www.python.org/dev/peps/pep-0440/#version-specifiers
+                #   See https://peps.python.org/pep-0440/#version-specifiers
                 docopt == 0.6.1             # Version Matching. Must be version 0.6.1
                 keyring >= 4.1.1            # Minimum version 4.1.1
                 coverage != 3.5             # Version Exclusion. Anything except version 3.5
@@ -338,15 +368,15 @@ def test_parse_requirements_stress(chroot):
                 green
                 #
                 """
-            ),
+            )
         )
-    )
+    req_iter = parse_requirement_file(root_requirements_txt)
 
     # Ensure local non-distribution files matching distribution names are not erroneously probed
     # as distributions to find name and version metadata.
-    touch("nose")
+    touch(tmpdir.join("nose"))
 
-    touch("downloads/numpy-1.9.2-cp34-none-win32.whl")
+    touch(tmpdir.join("downloads", "numpy-1.9.2-cp34-none-win32.whl"))
     with environment_as(PROJECT_NAME="Project"):
         results = normalize_results(req_iter)
 
@@ -366,7 +396,7 @@ def test_parse_requirements_stress(chroot):
         req(project_name="SomeProject", specifier="==5.4", marker="python_version < '2.7'"),
         req(project_name="SomeProject", marker="sys_platform == 'win32'"),
         url_req(project_name="SomeProject", url="https://example.com/somewhere/over/here"),
-        local_req(path=os.path.realpath("somewhere/over/here")),
+        local_req(path=tmpdir.join("somewhere", "over", "here")),
         req(project_name="FooProject", specifier=">=1.2"),
         vcs_req(
             vcs=VCS.Git,
@@ -374,30 +404,45 @@ def test_parse_requirements_stress(chroot):
             url="https://git.example.com/MyProject.git@da39a3ee5e6b4b0d3255bfef95601890afd80709",
         ),
         vcs_req(vcs=VCS.Git, project_name="MyProject", url="ssh://git.example.com/MyProject"),
-        vcs_req(vcs=VCS.Git, project_name="MyProject", url="file:///home/user/projects/MyProject"),
+        vcs_req(
+            vcs=VCS.Git,
+            project_name="MyProject",
+            url="file:///home/user/projects/MyProject#subdirectory=pkg_dir",
+        ),
         Constraint(DUMMY_LINE, Requirement.parse("AnotherProject")),
         local_req(
-            path=os.path.realpath("extra/a/local/project"),
+            path=tmpdir.join("extra", "a", "local", "project"),
             extras=["foo"],
             marker="python_full_version == '2.7.8'",
         ),
         local_req(
-            path=os.path.realpath("extra/another/local/project"),
+            path=tmpdir.join("extra", "another", "local", "project"),
             marker="python_version == '2.7.*'",
         ),
-        local_req(path=os.path.realpath("extra/another/local/project")),
-        local_req(path=os.path.realpath("extra")),
-        local_req(path=os.path.realpath("extra/tmp/tmpW8tdb_")),
-        local_req(path=os.path.realpath("extra/tmp/tmpW8tdb_"), extras=["foo"]),
         local_req(
-            path=os.path.realpath("extra/tmp/tmpW8tdb_"),
+            path=tmpdir.join("extra", "another", "local", "project"),
+            marker="python_version == '2.7.*'",
+        ),
+        local_req(path=tmpdir.join("extra", "another", "local", "project")),
+        local_req(path=tmpdir.join("extra")),
+        local_req(path=tmpdir.join("extra", "tmp", "tmpW8tdb_")),
+        local_req(path=tmpdir.join("extra", "tmp", "tmpW8tdb_"), extras=["foo"]),
+        local_req(
+            path=tmpdir.join("extra", "tmp", "tmpW8tdb_"),
             extras=["foo"],
             marker="python_version == '3.9'",
         ),
         vcs_req(
             vcs=VCS.Mercurial,
             project_name="AnotherProject",
-            url="http://hg.example.com/MyProject@da39a3ee5e6b",
+            url="http://hg.example.com/MyProject@da39a3ee5e6b#subdirectory=foo/bar",
+            extras=["more", "extra"],
+            marker="python_version == '3.9.*'",
+        ),
+        vcs_req(
+            vcs=VCS.Mercurial,
+            project_name="AnotherProject",
+            url="http://hg.example.com/MyProject@da39a3ee5e6b#subdirectory=foo/bar",
             extras=["more", "extra"],
             marker="python_version == '3.9.*'",
         ),
@@ -414,6 +459,12 @@ def test_parse_requirements_stress(chroot):
             specifier="==1.9.2",
             marker="python_version == '3.4.*' and sys_platform == 'win32'",
         ),
+        url_req(
+            project_name="numpy",
+            url="https://a/numpy-1.9.2-cp34-none-win32.whl",
+            specifier="==1.9.2",
+            marker="python_version == '3.4.*' and sys_platform == 'win32'",
+        ),
         vcs_req(vcs=VCS.Git, project_name="Django", url="https://github.com/django/django.git"),
         vcs_req(
             vcs=VCS.Git,
@@ -425,15 +476,30 @@ def test_parse_requirements_stress(chroot):
             project_name="Django",
             url="https://github.com/django/django.git@fd209f62f1d83233cc634443cfac5ee4328d98b8",
         ),
-        url_req(
-            project_name="Django",
-            url=os.path.realpath("extra/projects/django-2.3.zip"),
+        file_req(
+            project_name="django",
+            url=tmpdir.join("extra", "projects", "django-2.3.zip"),
+            specifier="==2.3",
+            marker="python_version>='3.10'",
+        ),
+        file_req(
+            project_name="django",
+            url=tmpdir.join("extra", "projects", "django-2.3.zip"),
             specifier="==2.3",
             marker="python_version>='3.10'",
         ),
         url_req(
+            project_name="torch",
+            url="http://download.pytorch.org/whl/cpu/torch-1.12.1%2Bcpu-cp310-cp310-linux_x86_64.whl",
+            specifier="==1.12.1+cpu",
+        ),
+        local_req(path=tmpdir.join("extra", "a", "local", "project"), editable=True),
+        local_req(path=tmpdir.join("extra", "a", "local", "project"), editable=True),
+        local_req(path=tmpdir.join("extra", "another", "local", "project"), editable=True),
+        local_req(path=tmpdir.join("extra", "another", "local", "project"), editable=True),
+        file_req(
             project_name="numpy",
-            url=os.path.realpath("./downloads/numpy-1.9.2-cp34-none-win32.whl"),
+            url=tmpdir.join("downloads", "numpy-1.9.2-cp34-none-win32.whl"),
             specifier="==1.9.2",
         ),
         url_req(
