@@ -1,4 +1,4 @@
-# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
+# Copyright 2020 Pex project contributors.
 # Licensed under the Apache License, Version 2.0 (see LICENSE).
 
 from __future__ import absolute_import
@@ -7,7 +7,6 @@ import errno
 import multiprocessing
 import os
 import shutil
-import subprocess
 import sys
 import tempfile
 from subprocess import CalledProcessError
@@ -15,16 +14,16 @@ from textwrap import dedent
 
 import pytest
 
-from pex.common import safe_mkdtemp, safe_open, temporary_dir, touch
+from pex.common import CopyMode, safe_mkdtemp, safe_open, temporary_dir, touch
 from pex.compatibility import PY2
 from pex.executor import Executor
 from pex.interpreter import PythonInterpreter
 from pex.layout import Layout
-from pex.pex_builder import CopyMode, PEXBuilder
-from pex.testing import IS_PYPY, PY310, PY_VER, ensure_python_interpreter, run_pex_command
+from pex.pex_builder import PEXBuilder
 from pex.typing import TYPE_CHECKING, cast
 from pex.util import named_temporary_file
 from pex.venv.virtualenv import Virtualenv
+from testing import IS_PYPY, PY310, PY_VER, ensure_python_interpreter, run_pex_command, subprocess
 
 if TYPE_CHECKING:
     from typing import Any, Dict, Iterable, Iterator, List, Optional, Protocol, Set, Text, Tuple
@@ -48,11 +47,20 @@ def pex():
         touch(os.path.join(src_dir, "user/__init__.py"))
         touch(os.path.join(src_dir, "user/package/__init__.py"))
 
+        # Fabric dips into Invoke vendored code. It depends on "invoke<2.0,>=1.3", but in version
+        # 1.7.0, the vendored `decorator` module Fabric depends on inside Invoke no longer is
+        # importable under Python 2.7; so we pin low.
+        constraints = os.path.join(tmpdir, "constraints.txt")
+        with open(constraints, "w") as fp:
+            fp.write("Invoke==1.6.0")
+
         # N.B.: --unzip just speeds up runs 2+ of the pex file and is otherwise not relevant to
         # these tests.
         run_pex_command(
             args=[
                 "fabric=={}".format(FABRIC_VERSION),
+                "--constraints",
+                constraints,
                 "-c",
                 "fab",
                 "--sources-directory",
@@ -61,14 +69,18 @@ def pex():
                 pex_path,
                 "--include-tools",
             ]
-        )
+        ).assert_success()
         yield os.path.realpath(pex_path)
 
 
 def make_env(**kwargs):
     # type: (**Any) -> Dict[str, str]
     env = os.environ.copy()
-    env.update((k, str(v)) for k, v in kwargs.items())
+    for k, v in kwargs.items():
+        if v is None:
+            env.pop(k, None)
+        else:
+            env[k] = str(v)
     return env
 
 
@@ -143,6 +155,14 @@ def parse_fabric_version_output(output):
     return dict(cast("Tuple[Text, Text]", line.split(" ", 1)) for line in output.splitlines())
 
 
+@pytest.mark.skipif(
+    sys.version_info[:2] >= (3, 12),
+    reason=(
+        "Fabric depends on invoke which embeds six which uses a meta path importer that only "
+        "implements the PEP-302 finder spec and not the modern spec. Only the modern finder spec "
+        "is supported by Python 3.12+."
+    ),
+)
 def test_venv_pex(create_pex_venv):
     # type: (CreatePexVenv) -> None
     venv = create_pex_venv()
@@ -250,6 +270,14 @@ def test_binary_path(create_pex_venv):
     assert 0 == returncode
 
 
+@pytest.mark.skipif(
+    sys.version_info[:2] >= (3, 12),
+    reason=(
+        "Fabric depends on invoke which embeds six which uses a meta path importer that only "
+        "implements the PEP-302 finder spec and not the modern spec. Only the modern finder spec "
+        "is supported by Python 3.12+."
+    ),
+)
 def test_venv_pex_interpreter_special_modes(create_pex_venv):
     # type: (CreatePexVenv) -> None
     venv = create_pex_venv()
@@ -452,6 +480,13 @@ def test_venv_copies(tmpdir):
     assert not os.path.islink(venv_copies_interpreter.binary)
 
 
+@pytest.mark.skipif(
+    PythonInterpreter.get().identity.pypy_version == (7, 3, 14),
+    reason=(
+        "The PyPy 7.3.14 release creates non-relocatable venvs: "
+        "https://github.com/pypy/pypy/issues/4838"
+    ),
+)
 def test_relocatable_venv(tmpdir):
     # type: (Any) -> None
 
@@ -534,13 +569,19 @@ def test_compile(tmpdir):
     venv = os.path.join(str(tmpdir), "venv")
     subprocess.check_call(args=[pex_file, "venv", venv], env=make_env(PEX_TOOLS=1))
     # N.B.: The right way to discover the site-packages dir is via site.getsitepackages().
-    # Unfortunately we use an old version of virtualenv to create PyPy and CPython 2.7 venvs and it
-    # does not add a getsitepackages function to site.py; so we cheat.
-    if IS_PYPY:
+    # Unfortunately we use an old version of virtualenv to create PyPy <= 3.7 and CPython 2.7 venvs
+    # and it does not add a getsitepackages function to site.py; so we cheat.
+    if IS_PYPY and PY_VER <= (3, 7):
         site_packages = "site-packages"
     else:
         site_packages = os.path.join(
-            "lib", "python{}.{}".format(sys.version_info[0], sys.version_info[1]), "site-packages"
+            "lib",
+            "{python}{major}.{minor}".format(
+                python="pypy" if IS_PYPY else "python",
+                major=sys.version_info[0],
+                minor=sys.version_info[1],
+            ),
+            "site-packages",
         )
 
     # Ensure we have at least the basic direct dependency python files we expect.
@@ -638,7 +679,7 @@ def test_strip_pex_env(tmpdir):
 def test_warn_unused_pex_env_vars():
     # type: () -> None
     # N.B.: We don't use the pytest tmpdir fixture here since it creates fairly length paths under
-    # /tmp and under macOS, where TMPDIR is already fairly deeply nested, we trigger Pex warinings
+    # /tmp and under macOS, where TMPDIR is already fairly deeply nested, we trigger Pex warnings
     # about script shebang length. Those warnings pollute stderr.
     tmpdir = safe_mkdtemp()
     venv_pex = os.path.join(tmpdir, "venv.pex")
@@ -653,7 +694,12 @@ def test_warn_unused_pex_env_vars():
         stdout, stderr = process.communicate()
         assert 0 == process.returncode
         assert not stdout
-        assert expected_stderr.strip() == stderr.decode("utf-8").strip()
+        expected = expected_stderr.strip()
+        error = stderr.decode("utf-8").strip()
+        if expected:
+            assert error.endswith(expected), error
+        else:
+            assert not error, error
 
     assert_execute_venv_pex(expected_stderr="")
     assert_execute_venv_pex(expected_stderr="", PEX_ROOT=os.path.join(tmpdir, "pex_root"))
@@ -661,6 +707,7 @@ def test_warn_unused_pex_env_vars():
     assert_execute_venv_pex(expected_stderr="", PEX_EXTRA_SYS_PATH="more")
     assert_execute_venv_pex(expected_stderr="", PEX_VERBOSE="0")
 
+    assert_execute_venv_pex(expected_stderr="", PEX_INHERIT_PATH="fallback")
     assert_execute_venv_pex(
         expected_stderr=dedent(
             """\
@@ -669,6 +716,7 @@ def test_warn_unused_pex_env_vars():
             """
         ),
         PEX_INHERIT_PATH="fallback",
+        PEX_VERBOSE="0",
     )
 
     assert_execute_venv_pex(
@@ -681,7 +729,7 @@ def test_warn_unused_pex_env_vars():
         ),
         PEX_COVERAGE="1",
         PEX_INHERIT_PATH="fallback",
-        PEX_VERBOSE="0",
+        PEX_VERBOSE="1",
     )
 
 
@@ -706,7 +754,7 @@ def test_custom_prompt(tmpdir):
         args=[venv_pex, "venv", "--prompt", "jane", venv_dir], env=make_env(PEX_TOOLS=True)
     )
 
-    if PY_VER == (2, 7) or IS_PYPY:
+    if PY_VER == (2, 7) or (IS_PYPY and PY_VER <= (3, 7)):
         # Neither CPython 2.7 not PyPy interpreters have (functioning) venv modules; so we create
         # their venvs with an old copy of virtualenv that does not surround the prompt with parens.
         expected_prompt = "jane"
@@ -738,7 +786,7 @@ def test_remove(
     # type: (...) -> None
     pex_root = os.path.join(str(tmpdir), "pex_root")
 
-    def create_venv_pex():
+    def create_pex():
         # type: () -> str
         venv_pex = os.path.join(str(tmpdir), "venv.pex")
         run_pex_command(
@@ -747,6 +795,8 @@ def test_remove(
                 pex_root,
                 "--runtime-pex-root",
                 pex_root,
+                "--layout",
+                str(layout),
                 "-o",
                 venv_pex,
                 "--include-tools",
@@ -757,29 +807,33 @@ def test_remove(
     venv_dir = os.path.join(str(tmpdir), "venv_dir")
     assert not os.path.exists(venv_dir)
 
-    venv_pex = create_venv_pex()
-    subprocess.check_call(args=[venv_pex, "venv", venv_dir], env=make_env(PEX_TOOLS=True))
+    pex = create_pex()
+    subprocess.check_call(
+        args=[sys.executable, pex, "venv", venv_dir], env=make_env(PEX_TOOLS=True, PEX_ROOT=None)
+    )
     assert os.path.exists(venv_dir)
-    assert os.path.exists(venv_pex)
+    assert os.path.exists(pex)
     assert os.path.exists(pex_root)
 
     shutil.rmtree(venv_dir)
     assert not os.path.exists(venv_dir)
 
     subprocess.check_call(
-        args=[venv_pex, "venv", "--rm", "pex", venv_dir], env=make_env(PEX_TOOLS=True)
+        args=[sys.executable, pex, "venv", "--rm", "pex", venv_dir],
+        env=make_env(PEX_TOOLS=True, PEX_ROOT=None),
     )
     assert os.path.exists(venv_dir)
-    assert not os.path.exists(venv_pex)
+    assert not os.path.exists(pex)
     assert os.path.exists(pex_root)
 
     shutil.rmtree(venv_dir)
     assert not os.path.exists(venv_dir)
-    venv_pex = create_venv_pex()
+    pex = create_pex()
 
     subprocess.check_call(
-        args=[venv_pex, "venv", "--rm", "all", venv_dir], env=make_env(PEX_TOOLS=True)
+        args=[sys.executable, pex, "venv", "--rm", "all", venv_dir],
+        env=make_env(PEX_TOOLS=True, PEX_ROOT=None),
     )
     assert os.path.exists(venv_dir)
-    assert not os.path.exists(venv_pex)
+    assert not os.path.exists(pex)
     assert not os.path.exists(pex_root)
